#include "../global.h"

#include <time.h>

// qtils
#include <Config.h>
#include <RandomUtils.h>
#include <TimeUtils.h>
#include <logger/Log.h>

// bass
#include <Bass.h>
#include <db/Database.h>

#include "SpectraTH.h"

SpectraTH::SpectraTH(Config *conf) : TaskHandler(conf){

}
SpectraTH::~SpectraTH() {
	// not my Config*
}

void SpectraTH::HandleTask() {
	string command = mConfig->Read<string>("command");

	if(command.compare("estimateNumPatterns") == 0)				 EstimateNumPatterns();
	else	if(command.compare("estimateNumExpected") == 0)		 EstimateNumExpectedPatternsIndependent();
	else	if(command.compare("estimateNumClosed") == 0)		 EstimateNumClosedPatterns();
	else	THROW(string("Unable to handle task " + command));
}

string SpectraTH::BuildWorkingDir() {
	string command = mConfig->Read<string>("command");

	if(command.compare("estimateNumPatterns") == 0  || command.compare("estimateNumPatternsRange") == 0  || command.compare("estimateNumExpected") == 0)
		return  command + "/" + mConfig->Read<string>("dbName") + "-" + Logger::GetTimestamp() + "/";
	if(command.compare("estimateNumClosed") == 0)
		return  command + "/" + mConfig->Read<string>("dbName") + "-" + mConfig->Read<string>("minsup") + "-" + Logger::GetTimestamp() + "/";

	return command + "/";
}

void SpectraTH::EstimateNumPatterns() {
	string dbName = mConfig->Read<string>("dbName");
	Database *db = Database::RetrieveDatabase(dbName);
	db->EnableFastDBOccurences();

	// Config
	uint32 maxNumEstimates = mConfig->Read<uint32>("numEstimates", 1000);
	bool fixedMinsup = mConfig->Read<bool>("fixedMinsup");
	bool outputSamples = mConfig->Read<bool>("outputSamples", false);
	uint32 minMinsup = mConfig->Read<uint32>("minMinsup");
	uint32 maxMinsup = mConfig->Read<uint32>("maxMinsup", db->GetNumRows());
	uint32 maxDepth = mConfig->Read<uint32>("maxDepth", db->GetAlphabetSize());

	// Seed
	uint32 seed = mConfig->Read<uint32>("seed", 0);
	if(seed == 0)
		RandomUtils::Init();
	else
		RandomUtils::Init(seed);

	// Init data structures
	size_t alphSize = db->GetAlphabetSize();
	uint32 *alphSupps = new uint32[alphSize];
	{
		alphabet *alph = db->GetAlphabet();
		for(alphabet::iterator it = alph->begin(); it!=alph->end(); it++)
			alphSupps[it->first] = it->second;
	}
	islist *level = new islist();
	maxMinsup = min(maxMinsup, alphSupps[0]); // the first item is the most frequent one
	
	// Open file
	string baseFilename = Bass::GetWorkingDir() + dbName + "-" + mConfig->Read<string>("minMinsup");
	if(!fixedMinsup) {
		char temp[10];
		sprintf_s(temp, 10, "%u", maxMinsup);
		baseFilename += "-" + string(temp);
	}
	baseFilename += "-" + Logger::GetTimestamp();
	string filename = baseFilename + "-estimates.txt";
	FILE *fp;
	if(!fixedMinsup || outputSamples)
		fopen_s(&fp, filename.c_str(), "w");

	// Init empty root candidate
	ItemSet *root = ItemSet::CreateEmpty(db->GetDataType(), db->GetAlphabetSize());
	Timer *timer = new Timer();

	// Start
	double estimateSum = 0;
	uint32 numEstimates = 0;
	uint32 minsup = minMinsup;
	uint64 numSupComputations = 0;

	while(numEstimates < maxNumEstimates) {
		// While candidates and within maxDepth
		ItemSet *current = root;
		double estimate = 0;
		double levelEstimate = 1;
		uint32 depth=1;

		// Randomize minCoverage?
		if(!fixedMinsup)
			minsup = minMinsup + RandomUtils::UniformUint32(maxMinsup - minMinsup + 1);

		for(; depth <= maxDepth && current != NULL; ++depth) {
			// Generate and test next level candidates
			//printf_s("Generating & testing level-%u candidates ...\r", depth);
			for(size_t i=0; i<alphSize; i++) {
				if(alphSupps[i] < minsup)
					break;
				if(current->IsItemInSet(i))
					continue;

				// Create candidate & compute support
				ItemSet *cand = db->ExtendAndDetermineOccurrences(current, i);
				++numSupComputations;

				// Add to next level if support exceeds threshold
				if(cand->GetSupport() >= minsup) {
					level->push_back(cand);
				} else {	// discard otherwise
					delete cand;
				}
			}

			// Update estimate
			levelEstimate *= level->size();
			levelEstimate /= depth;
			estimate += levelEstimate;

			// Cleanup
			if(current != root)
				delete current;

			// Pick one & discard rest
			if(depth < maxDepth && level->size()>0) {
				uint32 pickMe = RandomUtils::UniformUint32((uint32)level->size());
				islist::iterator it = level->begin();
				for(; pickMe>0; pickMe--) {
					delete *it;
					it++;
				}
				current = *it;
				it++;
				while(it != level->end()) {
					delete *it;
					it++;
				}
			} else {
				for(islist::iterator it = level->begin(); it != level->end(); it++)
					delete *it;
				current = NULL;
			}
			level->clear();
		}

		//sprintf_s(temp, 500, "%I64u", estimate);
		//LOG_MSG("** Estimate: " + string(temp) + "\t\t\t\n");

		estimateSum += estimate;
		numEstimates++;
		//printf_s("%u\r", numEstimates);

		// Need to output this?
		if(!fixedMinsup || outputSamples)
			fprintf_s(fp, "%u,%f\n", minsup, estimate);
	}

	double estimate = estimateSum / numEstimates;
	if(!fixedMinsup || outputSamples)
		fclose(fp);
	if(fixedMinsup) {
		printf_s("%.01f\n", estimate);
		printf_s("(average over %u estimates)\n", numEstimates);
	}

	// Write to file
	filename = baseFilename + ".txt";
	fopen_s(&fp, filename.c_str(), "w");
	if(fixedMinsup)
		fprintf_s(fp, "%s;%u;-;%s;%.01f;%u;%I64u;%u\n", dbName.c_str(), minMinsup, mConfig->Read<string>("maxDepth", string("-")).c_str(), estimate, numEstimates, numSupComputations, (uint32) timer->GetElapsedTime());
	else
		fprintf_s(fp, "%s;%u;%u;%s;%u;%I64u;%u\n", dbName.c_str(), minMinsup, maxMinsup, mConfig->Read<string>("maxDepth", string("-")).c_str(), numEstimates, numSupComputations, (uint32) timer->GetElapsedTime());
	fclose(fp);
	
	// Cleanup
	delete timer;
	delete[] alphSupps;
	delete level;
	delete root;
	delete db;
}

void SpectraTH::EstimateNumExpectedPatternsIndependent() {
	string dbName = mConfig->Read<string>("dbName");
	Database *db = Database::RetrieveDatabase(dbName);

	// Config
	uint32 minMinsup = mConfig->Read<uint32>("minMinsup");
	uint32 maxMinsup = min(mConfig->Read<uint32>("maxMinsup", db->GetNumRows()), db->GetNumRows());
	uint32 numEstimatesPerSup = mConfig->Read<uint32>("numEstimatesPerSup", 1);

	// Seed
	uint32 seed = mConfig->Read<uint32>("seed", 0);
	if(seed == 0)
		RandomUtils::Init();
	else
		RandomUtils::Init(seed);

	// Init data structures
	size_t alphSize = db->GetAlphabetSize();
	uint32 *alphSupps = new uint32[alphSize];
	float *alphProbs = new float[alphSize];
	{
		float numRows = (float) db->GetNumRows();
		alphabet *alph = db->GetAlphabet();
		for(alphabet::iterator it = alph->begin(); it!=alph->end(); it++) {
			alphSupps[it->first] = it->second;
			alphProbs[it->first] = it->second / numRows;
		}
	}
	islist *level = new islist();

	// Open file
	string baseFilename = Bass::GetWorkingDir() + dbName + "-" + mConfig->Read<string>("minMinsup");
	char temp[10];
	sprintf_s(temp, 10, "%u", maxMinsup);
	baseFilename += "-" + string(temp);
	baseFilename += "-" + Logger::GetTimestamp();
	string filename = baseFilename + "-estimates.txt";
	FILE *fp;
	fopen_s(&fp, filename.c_str(), "w");

	// Init empty root candidate
	ItemSet *root = ItemSet::CreateEmpty(db->GetDataType(), db->GetAlphabetSize());
	root->SetSupport(db->GetNumRows());
	Timer *timer = new Timer();

	// Start
	for(uint32 minsup = minMinsup; minsup < maxMinsup; minsup++) {
		for(uint32 num=0; num<numEstimatesPerSup; num++) {
			// While candidates and within maxDepth
			ItemSet *current = root;
			double estimate = 0;
			double levelEstimate = 1;

			for(uint32 depth=1; depth <= alphSize && current != NULL; ++depth) {
				uint32 currSupport = current->GetSupport();

				// Generate and test next level candidates
				for(size_t i=0; i<alphSize; i++) {
					if(alphSupps[i] < minsup)
						break;
					if(current->IsItemInSet(i))
						continue;

					uint32 support = (uint32)(currSupport * alphProbs[i]);

					// Add to next level if support exceeds threshold
					if(support >= minsup) {
						ItemSet *cand = current->Clone();
						cand->AddItemToSet(i);
						cand->SetSupport(support);

						level->push_back(cand);
					}
				}

				// Update estimate
				levelEstimate *= level->size();
				levelEstimate /= depth;
				estimate += levelEstimate;

				// Cleanup
				if(current != root)
					delete current;

				// Pick one & discard rest
				if(depth < alphSize && level->size()>0) {
					uint32 pickMe = RandomUtils::UniformUint32((uint32)level->size());
					islist::iterator it = level->begin();
					for(; pickMe>0; pickMe--) {
						delete *it;
						it++;
					}
					current = *it;
					it++;
					while(it != level->end()) {
						delete *it;
						it++;
					}
				} else {
					for(islist::iterator it = level->begin(); it != level->end(); it++)
						delete *it;
					current = NULL;
				}
				level->clear();
			}
			fprintf_s(fp, "%u,%f\n", minsup, estimate);
		}
	}
	fclose(fp);

	// Write summary to file
	filename = baseFilename + ".txt";
	fopen_s(&fp, filename.c_str(), "w");
	fprintf_s(fp, "%s;%u;%u;%s;%u\n", dbName.c_str(), minMinsup, maxMinsup, mConfig->Read<string>("maxDepth", string("-")).c_str(), timer->GetElapsedTime());
	fclose(fp);

	// Cleanup
	delete timer;
	delete[] alphSupps;
	delete[] alphProbs;
	delete level;
	delete root;
	delete db;
}

void SpectraTH::EstimateNumClosedPatterns() {
	string dbName = mConfig->Read<string>("dbName");
	Database *db = Database::RetrieveDatabase(dbName);
	db->EnableFastDBOccurences();

	// Config
	uint32 maxNumEstimates = mConfig->Read<uint32>("numEstimates", 100);
	float minsupFloat = mConfig->Read<float>("minsup");
	uint32 minsup = mConfig->Read<uint32>("minsup");
	uint32 maxDepth = mConfig->Read<uint32>("maxDepth", db->GetAlphabetSize());

	if(minsupFloat < 1.0f) // --> relative support
		minsup = (uint32)(db->GetNumRows() * minsupFloat);

	// Seed & timer
	uint32 seed = mConfig->Read<uint32>("seed", 0);
	if(seed == 0)
		RandomUtils::Init();
	else
		RandomUtils::Init(seed);
	Timer *timer = new Timer();

	// Init data structures
	size_t alphSize = db->GetAlphabetSize();
	islist *level = new islist();
	uint32 *alphSupps = new uint32[alphSize];
	{
		alphabet *alph = db->GetAlphabet();
		for(alphabet::iterator it = alph->begin(); it!=alph->end(); it++)
			alphSupps[it->first] = it->second;
	}
	ItemSet *root = ItemSet::CreateEmpty(db->GetDataType(), alphSize);

	uint64 *closedCounts = new uint64[maxDepth+1];
	uint64 *frequentCounts = new uint64[maxDepth+1];
	double *levelEstimates = new double[maxDepth+1];
	for(uint32 i=0; i<=maxDepth; i++) {
		closedCounts[i] = 0;
		frequentCounts[i] = 0;
		levelEstimates[i] = 0;
	}

	// Start
	double estimateSum = 0;
	uint32 numEstimates = 0;
	uint64 numSupComputations = 0;
	uint64 numClosureComputations = 0;

	while(numEstimates < maxNumEstimates) {
		// While candidates and within maxDepth
		ItemSet *current = root;
		double estimate = 0;
		double levelEstimate = 1;

		for(uint32 depth=1; depth <= maxDepth && current != NULL; ++depth) {
			//printf_s("Generating and testing level-%u candidates ...\r", depth);
			uint32 freqCount = 0;
			uint32 clsCount = 0;
			bool firstHit = true;

			// Generate and test all frequent extensions of current node
			for(size_t i=0; i<alphSize; i++) {
				if(alphSupps[i] < minsup)
					break;
				if(current->IsItemInSet(i))
					continue;

				// Create candidate & compute support
				ItemSet *cand = db->ExtendAndDetermineOccurrences(current, i);
				++numSupComputations;

				// Only consider if support exceeds threshold
				if(cand->GetSupport() >= minsup) {
					++freqCount; 

					// Determine whether closed or not
					++numClosureComputations;
					if(IsClosed(db, current, cand))
						++clsCount;

					// Add to next level regardless of closedness
					level->push_back(cand);
				} else {	// discard otherwise
					delete cand;
				}
			}

			frequentCounts[depth] += freqCount;
			closedCounts[depth] += clsCount;

			// Update estimate
			levelEstimate *= level->size();
			levelEstimate /= depth;
			levelEstimates[depth] += levelEstimate;
			estimate += levelEstimate;

			// Cleanup (but leave root be)
			if(current != root)
				delete current;

			// Pick one & discard rest
			if(depth < maxDepth && level->size()>0) {
				uint32 pickMe = RandomUtils::UniformUint32((uint32)level->size());
				islist::iterator it = level->begin();
				for(; pickMe>0; pickMe--) {
					delete *it;
					it++;
				}
				current = *it;
				it++;
				while(it != level->end()) {
					delete *it;
					it++;
				}
			} else {
				for(islist::iterator it = level->begin(); it != level->end(); it++)
					delete *it;
				current = NULL;
			}
			level->clear();
		}

		//sprintf_s(temp, 500, "%I64u", estimate);
		//LOG_MSG("** Estimate: " + string(temp) + "\t\t\t\n");

		estimateSum += estimate;
		numEstimates++;
		// printf_s("%.01f\r", estimateSum / numEstimates);
	}

	// Output final result: estimates for 1) all and 2) closed frequent itemset counts
	double estimate = estimateSum / numEstimates;
	double closedEstimate = 0;
	for(uint32 d=1; d<=maxDepth; ++d) {
		if(frequentCounts[d]==0)
			continue;
		closedEstimate += ((double)closedCounts[d] / frequentCounts[d]) * levelEstimates[d];
		//printf_s("%f\n", levelEstimates[d]);
	}
	closedEstimate /= numEstimates;

	printf_s("Frequent:\t%.01f\n", estimate);
	printf_s("Closed:\t\t%.01f\n", closedEstimate);
	printf_s("(average over %u estimates)\n", numEstimates);

	// Write to file
	string filename = Bass::GetWorkingDir() + dbName + "-" + mConfig->Read<string>("minsup") + "-" + Logger::GetTimestamp() + ".txt";
	FILE *fp;
	fopen_s(&fp, filename.c_str(), "w");
	fprintf_s(fp, "%s;%u;%s;%.01f;%.01f;%u;%I64u;%I64u;%u\n", dbName.c_str(), minsup, mConfig->Read<string>("maxDepth", string("-")).c_str(), estimate, closedEstimate, numEstimates, numSupComputations, numClosureComputations, timer->GetElapsedTime());
	fclose(fp);

	// Cleanup
	delete timer;
	delete level;
	delete[] alphSupps;
	delete[] closedCounts;
	delete[] frequentCounts;
	delete[] levelEstimates;
	delete root;
	delete db;
}

bool SpectraTH::IsClosed(Database *db, ItemSet *parent, ItemSet *cand) {
	ItemSet **rows = db->GetRows();
	uint32 *tidList = cand->GetDBOccurrence();
	uint32 numOccs = cand->GetNumDBOccurs();

	ItemSet *closed = rows[tidList[0]]->Clone();
	for(uint32 r=1; r<numOccs && closed->GetLength()>cand->GetLength(); r++)
		closed->Intersect(rows[tidList[r]]);

	bool isClosed = cand->GetLength()==closed->GetLength();
	delete closed;

	return isClosed;
}
